package io.github.null2264.tsukumogami.core

import co.touchlab.kermit.Logger
import dev.kord.core.Kord
import dev.kord.core.entity.Message
import dev.kord.core.event.message.MessageCreateEvent
import dev.kord.gateway.Intent
import dev.kord.gateway.PrivilegedIntent
import io.github.null2264.tsukumogami.core.commands.Command
import io.github.null2264.tsukumogami.core.exceptions.CommandException
import io.github.null2264.tsukumogami.core.exceptions.CommandNotFound
import io.github.null2264.tsukumogami.core.module.BotModule
import io.github.null2264.tsukumogami.core.commands.IGroup
import io.github.null2264.tsukumogami.core.koin.TsukumogamiKoinComponent
import org.koin.core.component.inject

open class Bot internal constructor(): IGroup, TsukumogamiKoinComponent {

    val client: Kord by inject()
    private val modules = mutableMapOf<String, BotModule>()
    private val _prefixes = mutableListOf<String>()
    override val allCommands: MutableMap<String, Command> = mutableMapOf()

    suspend fun self() = client.getSelf().asUser()

    fun addModule(module: BotModule) {
        modules[module.name] = module.install(this)
    }

    fun addPrefix(prefix: String) {
        _prefixes.add(prefix)
    }

    private fun addCommand(name: String, command: Command) {
        if (allCommands.containsKey(name)) {
            throw IllegalStateException("Duplicate command: '${command.name}'")
        }
        allCommands[command.name] = command
    }

    fun addCommand(command: Command) {
        addCommand(command.name, command)
        command.alias.forEach { addCommand(it, command) }
    }

    private fun removeCommand(name: String) {
        if (!allCommands.containsKey(name)) {
            throw IllegalStateException("Command not found: '${name}'")
        }
        allCommands.remove(name)
    }

    fun removeCommand(command: Command) {
        removeCommand(command.name)
        command.alias.forEach { removeCommand(it) }
    }

    open suspend fun start() {
        client.login {
            @OptIn(PrivilegedIntent::class)
            intents += Intent.MessageContent
        }
    }

    open suspend fun stop() {
        client.shutdown()
    }

    fun getCommand(name: String?) = allCommands[name]

    private suspend fun getContext(message: Message): Context {
        val parsed = message.content.parsePrefixAndCommand()
        val context = Context(message, parsed?.first)
        context.command = getCommand(parsed?.second)
        if (context.command != null) context.invokedWith = parsed?.second
        return context
    }

    open suspend fun onCommandError(context: Context, error: CommandException) {
        context.send(error.message!!)
    }

    open suspend fun processCommand(message: Message) {
        val ctx = getContext(message)

        try {
            ctx.command?.invoke(ctx) ?: throw CommandNotFound()
        } catch (e: CommandException) {
            onCommandError(ctx, e)
        } catch (e: Exception) {
            Logger.e(e) { "Something went wrong while trying to process command '${ctx.command?.name}'" }
        }
    }

    open suspend fun onMessage(message: Message, event: MessageCreateEvent) {
        if (message.author?.isBot != false) return

        processCommand(message)
    }

    open suspend fun onReady() {
        Logger.i { "Online! ${client.getSelf().username}" }
    }

    open suspend fun getPrefixes(): List<String> {
        return _prefixes.toList()
    }

    private suspend fun String.parsePrefixAndCommand(): Pair<String, String>? {
        if (this.isBlank()) return null

        var ret: Pair<String, String>? = null

        run {
            getPrefixes().forEach { candidate ->
                val prefix = this.substring(0, candidate.length)
                if (prefix != candidate) { return@forEach }
                val command = this.drop(candidate.length).substringBefore(' ')
                ret = prefix to command
                return@run
            }
        }

        return ret
    }
}
